#include <Thread.h>
#include <UserThread.h>
#include <pthread.h>
#include <list>
#include <ProcessingUnit.h>
#include <iostream>
#include "affinity.h"
#include "globals.h"

using namespace std;

/**
 * A logical processing unit. the unit can be a CPU, another remote computer,
 * another thread, another Process ect'.
 * <p>
 * The processing unit execute all objects of classes that implement
 * {@link java.lang.Runnable} interface, that have been posted in it's
 * {@link #userActionsToExecute} queue, by executing it's
 * {@link java.lang.Runnable#run()} method.<br>
 * <p>
 * Each processing unit holds a thread to execute all runnable method. same
 * thread is used for executing all the runnables, I.E. no overhead for creating
 * new threads.
 * <p>
 * join methods are implemented here, in-order to block one thread (the thread
 * that called the join methods) until the processing unit finishes executing
 * the pending runnables, or a specific runnable.
 * <p>
 * the runnable objects are called userActions.
 *
 * for more information it is recomended that the code itself be viewd.the code
 * is self explanitory.
 *
 * @see Runnable
 * @author Gil Kulish
 * @author Eliah Ninyo
 */
ProcessingUnit::ProcessingUnit() {
	pthread_mutex_init(&pUnitObjectMutex ,NULL);
	pthread_cond_init(&pUnitObjectCond ,NULL);
	userActionsToExecute = new list<Runnable*> ();
	t = new Thread(this);
	// set as daemon thread, so that the program will end even if the thread
	// still executes
	t->start();

}
ProcessingUnit::~ProcessingUnit() {
}

/**
 * Inner class that contains the runnable method that defines the behaviour
 * of ProcessingUnit's daemon thread. Generally speaking, the thread blocks
 * until a <tt> Runnable </tt> object is inserted to it's
 * {@link #userActionsToExecute} queue, then it awakes, executes all
 * <tt> Runnable </tt> available in the queue and returens to wait again.
 */
void ProcessingUnit::run() {
	unique_processor_number_sequance++;
	/*set the affinity of a processing unit to a real processor (or core)*/
#ifndef NO_AFF
	setThreadAff(((unique_processor_number_sequance-1) % procnum)  +1);
#endif
	//pth_init must be called for each kernel thread from within the thread
	pth_init();
	tld = getPthThreadLocal();
	pthread_mutex_lock(&pUnitObjectMutex);
	/**
	 * gil kulish - todo - enable this for debug purpose

	 Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
	 Thread.currentThread().setName("Processor " + processorNumber);
	 */
	processorNumber = unique_processor_number_sequance;
	debug_printf3("processing unit - %d launched; pth thread local sched - %x:\n",
			processorNumber, tld->localPth_sched);
	try {
		for (;;) {
			while (userActionsToExecute->size() > 0) {
				Runnable* r = userActionsToExecute->front();
				userActionsToExecute->pop_front();
				/*	synchronized (r) {
				 */
				UserThread *ut = new UserThread(r);
				// when ut->run() is active; the lazy thread should be opened
				// using lazyThread->start() and not by pu->execute(lazyThread);
//				ut->run();
				//open a new user thread (ut->start) is on remark:
				ut->start();
				//				printf("actually executing some task\n");
				debug_printf2("processing unit %d had launched a new user thread\n",
						processorNumber);
				/*r.notifyAll();
				 }*/
			}
			debug_printf2("stage A on %d \n", processorNumber);
			// finished executing all userActions in list, notify
			pthread_cond_broadcast(&pUnitObjectCond);

			//unlock while waiting for pth threads to finish, in order to enable insertion of new userThreads
			pthread_mutex_unlock(&pUnitObjectMutex);

			debug_printf2("stage B on %d \n", processorNumber);

			//before we block in a way that does not enable any pth thread to run;/
			//we need to make sure that all pth threads are finished
			pth_yield(NULL);

			debug_printf2("stage C on %d \n", processorNumber);
			/* this pthread lock does not work OK with the PTH locks, that is the reason why we
			 * have to make sure that by this instruction there are no more waiting PTH threads */
			pthread_mutex_lock(&pUnitObjectMutex);

			debug_printf2("stage D on %d \n", processorNumber);

			if (!pth_check_no_threads_left()) {// if there are still pth threads; we cannot call to pthread_cond_wait,
				// so we will continue the loop
				debug_print_pth_threads_left();
				continue;
			}
			debug_printf2("stage E on %d \n", processorNumber);
			//there is a chance new threads were added so check it:
			if (userActionsToExecute->size() > 0)
				continue;

			debug_printf2("stage F on %d \n", processorNumber);

			// release lock, wait until a new action to perform is
			// added to the list; this wait will block and not enable any pth_thread to run in parallel
			pthread_cond_wait(&pUnitObjectCond, &pUnitObjectMutex);
		}
	} catch (exception &why) {
		cerr << "Caught exception " << why.what() << endl;
		cerr << "We could go on if we wanted" << endl;
		cerr << "Although in this demo we'll exit nice" << endl;
	}
	pthread_mutex_unlock(&pUnitObjectMutex);

}

/**
 * adds a runnable object <tt>r</tt> to the ProcessingUnit's
 * {@link #userActionsToExecute} queue. and executes it when all other runnables
 * entered before <tt>r</tt> finished executing (fifo queue)
 *
 * @param r
 *            the runnable object to be executed
 */
void ProcessingUnit::execute(Runnable *r) {
	//printf("execute was called on some task (PU %d)\n",processorNumber);
	pthread_mutex_lock(&pUnitObjectMutex);
	userActionsToExecute->push_back(r);
	// notify the processing unit to execute the new added action
	pthread_cond_broadcast(&pUnitObjectCond);
	pthread_mutex_unlock(&pUnitObjectMutex);
	//	printf("exiting execute on some task(PU %d)\n",processorNumber);

}

/**
 * wait for a certain action to finish executing (block until action is
 * finished)
 */
void ProcessingUnit::join(Runnable* r) {
	/*Gil Kulish - stub - need to add*/
}

/**
 * wait for all userActions to finish executing (block until all userActions are
 * finished)
 */
void ProcessingUnit::join() {
	/*Gil Kulish - stub - need to add*/
	pthread_mutex_lock(&pUnitObjectMutex);
	while (userActionsToExecute->size() > 0) {

		pthread_cond_wait(&pUnitObjectCond, &pUnitObjectMutex);
	}
	pthread_mutex_unlock(&pUnitObjectMutex);
}

pthread_cond_t ProcessingUnit::pUnitClassCond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t ProcessingUnit::pUnitClassMutex = PTHREAD_MUTEX_INITIALIZER;

int ProcessingUnit::unique_processor_number_sequance = 0;
bool ProcessingUnit::pth_check_no_threads_left() {
	return !(pth_pqueue_elements(tld->localPth_RQ)>0 || (pth_pqueue_elements(tld->localPth_WQ)>0) || (pth_pqueue_elements(tld->localPth_SQ)>0) || (pth_pqueue_elements(tld->localPth_NQ)>0));
}

void ProcessingUnit::debug_print_pth_threads_left() {
	if (pth_pqueue_elements(tld->localPth_RQ)>0)
		debug_printf3("RQ number of threads - %d in processing unit %d\n", pth_pqueue_elements(tld->localPth_RQ),processorNumber);
	if (pth_pqueue_elements(tld->localPth_WQ)>0)
		debug_printf3("WQ number of threads - %d in processing unit %d\n", pth_pqueue_elements(tld->localPth_WQ),processorNumber);
	if (pth_pqueue_elements(tld->localPth_SQ)>0)
		debug_printf3("SQ number of threads - %d in processing unit %d\n", pth_pqueue_elements(tld->localPth_SQ),processorNumber);
	if (pth_pqueue_elements(tld->localPth_NQ)>0)
		debug_printf3("NQ number of threads - %d in processing unit %d\n", pth_pqueue_elements(tld->localPth_NQ),processorNumber);
}

/** depricated */
void ProcessingUnit::pth_join_all() {
	//	printf("waiting for all pth threads to finish\n");
	// this variable is needed since there is a chance that after finishing with one queue, another queue
	// is now not empty
	void *rval;
	bool potentiallyNotEmpty = true;
	while (potentiallyNotEmpty) {
		potentiallyNotEmpty = false;

		while (pth_pqueue_elements(tld->localPth_RQ)>0) {
			pth_t ptht = pth_pqueue_tail(tld->localPth_RQ);
			if (ptht)
				pth_join(ptht, &rval);
			potentiallyNotEmpty = true;
		}
		while (pth_pqueue_elements(tld->localPth_WQ)>0) {
			pth_t ptht = pth_pqueue_tail(tld->localPth_WQ);
			if (ptht)
				pth_join(ptht, &rval);
			potentiallyNotEmpty = true;
		}
		while (pth_pqueue_elements(tld->localPth_SQ)>0) {
			pth_t ptht = pth_pqueue_tail(tld->localPth_SQ);
			if (ptht)
				pth_join(ptht, &rval);
			potentiallyNotEmpty = true;
		}
		while (pth_pqueue_elements(tld->localPth_NQ)>0) {
			pth_t ptht = pth_pqueue_tail(tld->localPth_NQ);
			if (ptht)
				pth_join(ptht, &rval);
			potentiallyNotEmpty = true;
		}
	}
	//	printf("all pth threads are finished\n");
}
