/**
 * Manages creation and use of the processing units. contains a pool of
 * processing units. gives a direct access to a processing unit, each time a job
 * should be executed, a specific processing unit can be chosen from the pool.
 *
 */
#include <vector>
#include <ProcessingUnit.h>
#include <ProcessorsManager.h>
#include <Thread.h>
#include <pth.h>
#include <map>
#include <string.h>
#include <ParForJob.h>
using namespace std;

struct cmp_str
{
   bool operator()(char const *a, char const *b)
   {
      return strcmp(a, b) < 0;
   }
};

extern map<char*,int,cmp_str> threadToProcessor;

extern void loadSchedule();
/**
 * adds a runnable object <tt>r</tt> to the ProcessingUnit's
 * {@link #userActionsToExecute} queue. and executes it when all other runnables
 * entered before <tt>r</tt> finished executing (fifo queue)
 *
 * @param r
 *            the runnable object to be executed
 */
void ProcessorsManager::execute(Runnable *r) {
	static unsigned int processorsIter = 0;
//	printf("inside PM execute (number of scheduler rules %d)\n",threadToProcessor.size());
	if (threadToProcessor.size() > 0) {
		UserThread * ut = dynamic_cast<UserThread *> (r);
		if (ut) {
//			printf("after user thread\n");
			ParForJob * pfj = dynamic_cast<ParForJob *> (ut->toRun);
			if (pfj) {
				int pnum = threadToProcessor[pfj->threadLogicalName];
				if (pnum > 0) {
					processorsIter = pnum-1;
					#ifdef DEBUG
					printf("using processor %d for thread %s\n", pnum,
							pfj->threadLogicalName);
					#endif
				}
			}
		}
	}
	ProcessingUnit *pu = getProcessor(processorsIter++);
	pu->execute(r);

}

vector<ProcessingUnit*> *ProcessorsManager::getProcessors() {
	/*Gil Kulish - todo - need to synchronize*/
	if (!processors)
		processors = new vector<ProcessingUnit*> ();
	return processors;
}

/**
 * gets a specific ProcessingUnit <tt> n </tt> if the chosen ProcessingUnit
 * number is out of bounds <code>(n>=activeProcessorsNumber)</code>, it
 * returns the <code>n % activeProcessorsNumber</code> ProcessingUnit.
 */
ProcessingUnit *ProcessorsManager::getProcessor(int n) {

	return (*getProcessors())[n % activeProcessorsNumber];

}

/**
 * number of {@link ProcessingUnits} been created and can be used (a pool)
 * only {@link activeProcessorsNumber} will be used.
 */
/* synchronized */void ProcessorsManager::addProcessor() {
	activeProcessorsNumber++;
	if (numberOfProcessorsInPool < activeProcessorsNumber) {
		ProcessingUnit *pu = new ProcessingUnit();
		getProcessors()->push_back(pu);
		numberOfProcessorsInPool++;
	}
}

/* synchronized */
void ProcessorsManager::addNProcessors(int n) {
	for (int i = 0; i < n; i++) {
		addProcessor();
	}
}

/* synchronized */
void ProcessorsManager::setNProcessors(int n) {
	if (n > activeProcessorsNumber)
		for (int i = activeProcessorsNumber; i < n; i++) {
			addProcessor();
		}
	else
		activeProcessorsNumber = n;
}

/* synchronized */
void ProcessorsManager::removeProcessor() {
	activeProcessorsNumber--;
}

ProcessorsManager::ProcessorsManager() {
	activeProcessorsNumber = DEFAULT_ACTIVE_PROCESSES_NUMBER;
	numberOfProcessorsInPool = DEFAULT_ACTIVE_PROCESSES_NUMBER * 2;
	for (int i = 0; i < numberOfProcessorsInPool; i++) {
		getProcessors()->push_back(new ProcessingUnit());
	}
	// creating a bigger pool, adding more processing units then what the
	// user declared he needs, so the performance
	// will not degragate when the user needs more ProcessingUnits
	// load the scheduler if there is any:
	loadSchedule();
}

ProcessorsManager::~ProcessorsManager() {
}
/**
 * as in {@link Thread#join()} blocks until all jobs of all processing units
 * that have been posted up to this moment, have finished executing.
 *
 * @see Thread#join()
 */
void ProcessorsManager::join() {
	for (int i = 0; i < activeProcessorsNumber; i++) {
		(*processors)[i]->join();
	}
}

/**
 * singleton
 */
ProcessorsManager *ProcessorsManager::getInstance() {
	/*Gil Kulish - todo - need to synchronize*/
	if (instance == NULL) {
		instance = new ProcessorsManager();
	}
	return instance;
}

/* synchronized */
int ProcessorsManager::getActiveProcessorsNumber() {
	return activeProcessorsNumber;
}

const int ProcessorsManager::DEFAULT_ACTIVE_PROCESSES_NUMBER = procnum;
ProcessorsManager *ProcessorsManager::instance = NULL;
vector<ProcessingUnit*> *ProcessorsManager::processors = NULL;
